/**
 * Quad Sim v0.21
 * Copyright (C) 2005  Giliam de Carpentier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * You can contact me at giliam@decarpentier.nl for questions concerning this license.
 */
#include <windows.h>
#include <mmsystem.h>
#include <conio.h>
#include "common.h"

GRAVITY gravity(9.81);
MAP map("mapfile.map");

UINT lastTick;
float timerQueue = 0.0f;
int cnt = 0;

V vCamera, vCameraTo, vCameraRot, vCameraRotDest, vCameraRel, vCameraRelDest;

MB *pBlock;
QUAD *pquad;

MP *pmp[8];
SPRINGBP *pspring[8];

bool launch, leftButton, middleButton, rightButton;
int lastLaunch = 0;
float motorm, handbrake, throttle;
float steerAng, steerAngDest;
V vDisp;
int windowW, windowH;

void display() {
}

void render() {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);	
	glLoadIdentity();
	gluLookAt(vCamera.x, vCamera.y, vCamera.z,
			  vCameraTo.x, vCameraTo.y, vCameraTo.z,
			  0.0f, 1.0f, 0.0f);

	map.render(vCamera);

	V vP[8];
	for (int i = 0; i < 8; i++)
		vP[i] = pspring[i]->pmbA->m2w(pspring[i]->vA);

	glBegin(GL_QUADS);
		glColor3f(0.8f,0.8f,0.0f);
		glVertex3fv(vP[1].p); glVertex3fv(vP[0].p); glVertex3fv(vP[2].p); glVertex3fv(vP[3].p);
		glVertex3fv(vP[4].p); glVertex3fv(vP[5].p); glVertex3fv(vP[7].p); glVertex3fv(vP[6].p);

		glColor3f(0.0f,0.8f,0.8f);
		glVertex3fv(vP[2].p); glVertex3fv(vP[0].p); glVertex3fv(vP[4].p); glVertex3fv(vP[6].p);
		glVertex3fv(vP[5].p); glVertex3fv(vP[1].p); glVertex3fv(vP[3].p); glVertex3fv(vP[7].p);

		glColor3f(0.8f,0.0f,0.8f);
		glVertex3fv(vP[0].p); glVertex3fv(vP[1].p); glVertex3fv(vP[5].p); glVertex3fv(vP[4].p);
		glVertex3fv(vP[3].p); glVertex3fv(vP[2].p); glVertex3fv(vP[6].p); glVertex3fv(vP[7].p);
	glEnd();

	for (i = 0; i < 8; i++)
		vP[i] = (i & 1 ? pquad->pfssL : pquad->pfssR)->parm->ppvvA->vP + pquad->pfssL->rmCaster * 
			V(0, (i & 2 ? .32f : -.1f), (i & 4 ? .30f : -.12f));

	glBegin(GL_QUADS);
		glColor3f(0.6f,0.6f,0.0f);
		glVertex3fv(vP[1].p); glVertex3fv(vP[0].p); glVertex3fv(vP[2].p); glVertex3fv(vP[3].p);
		glVertex3fv(vP[4].p); glVertex3fv(vP[5].p); glVertex3fv(vP[7].p); glVertex3fv(vP[6].p);

		glColor3f(0.0f,0.6f,0.6f);
		glVertex3fv(vP[2].p); glVertex3fv(vP[0].p); glVertex3fv(vP[4].p); glVertex3fv(vP[6].p);
		glVertex3fv(vP[5].p); glVertex3fv(vP[1].p); glVertex3fv(vP[3].p); glVertex3fv(vP[7].p);

		glColor3f(0.6f,0.0f,0.6f);
		glVertex3fv(vP[0].p); glVertex3fv(vP[1].p); glVertex3fv(vP[5].p); glVertex3fv(vP[4].p);
		glVertex3fv(vP[3].p); glVertex3fv(vP[2].p); glVertex3fv(vP[6].p); glVertex3fv(vP[7].p);
	glEnd();


	pquad->render();

	glFlush ();
	glutSwapBuffers ();

}

void idle() {
	UINT currTick = timeGetTime();
	timerQueue += (currTick - lastTick) * .001f;
	lastTick = currTick;

	if (timerQueue >= TAU) {
		for (; timerQueue >= TAU; timerQueue -= TAU) {

			int i;
			for (i = 0; i < 8; i++) {
				pspring[i]->addForce(pspring[i]->force());
				pmp[i]->addForce(gravity.force(pmp[i]));
				pmp[i]->addForce(map.force(pmp[i]));
			}

			pBlock->addForce(gravity.force(pBlock));

///
			if (launch) {
				lastLaunch = -2.f/TAU;
				launch = false;
			}

			if (lastLaunch++ < 0) {
				pBlock->addForce(pBlock->m2w(V(0,.9f,-.2f)),-gravity.force(pBlock)*1.15);
				pBlock->addMoment(pBlock->rmPhi * (-200 * pBlock->vOmega));
			}
			pBlock->addMoment(pBlock->rmPhi * (-1 * pBlock->vOmega));


			float throttleDest = 0.f;
			if (leftButton) throttleDest += 1.f;
			if (middleButton) throttleDest -= .5f;
			if (rightButton) FILTER(handbrake, 1.f, .99); else handbrake = 0.f;
			if (throttle < throttleDest && throttleDest > 0) throttle = throttleDest;
			if (throttle > throttleDest && throttleDest < 0) throttle = throttleDest;
			FILTER(throttle, throttleDest, .992);

			float s = MAX(28.f, -pquad->pbss->paxle->relOmega);
			V vM = throttle * (pBlock->rmPhi * V(-0.12*s*s+0.88*s+1000-14000/s,0,0));
			pBlock->addMoment(.5f * vM); 
			pquad->pbss->paxle->addMoment(-vM);
			pquad->pbss->pbearing->addMoment(.5f * vM);

///


			pquad->pre();

			FILTER(vDisp, pquad->pmb->vV * 3.6f, .95f);
			if (!(cnt++ & 0xFF)) {
				printf("%f km/h\n",vDisp.len());
			}

			for (i = 0; i < 8; i++)
				pmp[i]->advance();
			pquad->advance();

			pquad->post();

			FILTER(steerAng, steerAngDest, 0.96f);
			FILTER(vCameraTo, pBlock->vP, 0.97f);
			FILTER(vCameraRot, vCameraRotDest, 0.995f);
			FILTER(vCameraRel, vCameraRelDest, 0.996f);
			vCameraRelDest =  RM(pBlock->rmPhi.vector()*V(0,1,0)) * RM(vCameraRot * V(-1,0,0)) * V(0.0f, 0.0f, 2.f);
			float height = map.heightAt(vCamera.x, vCamera.z);
			V vCameraDest = vCameraTo + RM((vCameraRot) * V(0,1,0)) * vCameraRel;
			vCameraDest.y = MAX(vCameraDest.y, height + .3f);
			FILTER(vCamera, vCameraDest, 0.97f);
			vCamera.y = MAX(vCamera.y, height + .1f);


		}
		render();
		if (timerQueue < TAU)
			Sleep(1000 * TAU);
	}
}

void reshape(int w, int h) {
	windowW = w; windowH = h;
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, (float)w/h, 0.1f, 50.0f);
}

void special(int key, int x, int y) {
	if (key == GLUT_KEY_UP) vCameraRotDest.x += 0.15f;
	if (key == GLUT_KEY_DOWN) vCameraRotDest.x -= 0.15f;
	if (key == GLUT_KEY_LEFT) vCameraRotDest.y -= 0.15f;
	if (key == GLUT_KEY_RIGHT) vCameraRotDest.y += 0.15f;
	if (key == GLUT_KEY_END) exit(0);
	if (key == GLUT_KEY_HOME) launch = true;
	vCameraRotDest.x = MIN(1.5f, MAX(0.0f, vCameraRotDest.x));
}

void mouse(int button, int state, int x, int y) {
	if (button == GLUT_LEFT_BUTTON) leftButton = state == GLUT_DOWN;
	if (button == GLUT_MIDDLE_BUTTON) middleButton = state == GLUT_DOWN;
	if (button == GLUT_RIGHT_BUTTON) rightButton = state == GLUT_DOWN;
}

void motion(int x, int y) {
	float relX = (0.5f - (float)x / windowW) * 1.5;
	steerAngDest = MIN(1.0f, MAX(-1.0f, relX)) * 0.85f;
}

void main (int argc, char *argv[]) {
	vCamera = V(0,0,-10.f); vCameraTo = VZERO; vCameraRot = vCameraRotDest = V(.3f,0,0);

	pBlock = new MB(PVV(V(1,1.6f,0.f),V(0,0,0)),
			        PVV(V(0,HALFPI,0),V(0,0,0)),
			        230, V(20,20,15));// V(30,25,30));

	for (int i = 0; i < 8; i++) {
		V v = (V((i&1)*2, (i&2)*1.5, (i&4)/2.f) - V(1.f, 1.5f, 1.5f)) * .15;
		if (v.y > 0) v *= V(3.5f, 1.f, (v.z < 0 ? 2.f : 8.f));
		v.y -= .02f; v.z -= .02f;
		pmp[i] = new MP(&pBlock->m2pvv(v), 3.f);
		pspring[i] = new SPRINGBP(pBlock, v, pmp[i], 0.1f, 0.99f);
	}

	pquad = new QUAD(pBlock);

	printf ("initialising GLUT ");

	glutInit (&argc, argv);
	glutInitWindowSize (320, 240);
    glutInitWindowPosition (0,0);
    glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutCreateWindow ("Quad sim v0.21.  (C) 2005, Giliam de Carpentier.  Released under the GNU GPL");
	glClearColor(0.90f,0.95f,.98f,1);

	glEnable(GL_LINE_WIDTH); 
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glLineWidth(4);
	glPointSize(4);
//	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	lastTick = timeGetTime();

	glutDisplayFunc			(display);
	glutSpecialFunc			(special);
	glutReshapeFunc			(reshape);
	glutMouseFunc			(mouse);
	glutMotionFunc			(motion);
	glutPassiveMotionFunc	(motion);
	glutIdleFunc			(idle);

	printf ("[done]\n");

    glutMainLoop ();
}
  
